import requests
import yaml
import os
from typing import Optional, Dict, Any, List

# Import configuration
from .config import BASE_URL

# Determine the directory where this script is located
_SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
_SCHEMA_PATH = os.path.join(_SCRIPT_DIR, 'schema.yaml')


class CarapisClientError(Exception):
    """Custom exception for CarapisClient errors."""
    pass


class CarapisClient:
    """
    Python client for the Carapis Encar v2 API.
    Loads API definitions from schema.yaml.
    Uses BASE_URL defined in config.py.
    """

    def __init__(self, api_key: str):
        """Initialize the client.

        Args:
            api_key: Your Carapis API key.
        """
        if not api_key:
            raise ValueError("api_key cannot be empty.")

        # Use BASE_URL from config, ensure no trailing slash
        self.base_url = BASE_URL.rstrip('/')
        self.api_key = api_key
        # Define the specific API path part we are interested in
        self.api_base_path = "/api/encar/v2"
        self._headers = {
            "Accept": "application/json",
            "Authorization": f"ApiKey {self.api_key}",
            "User-Agent": f"CarapisClientPython/{self._get_version()}"  # Include version
        }

        try:
            with open(_SCHEMA_PATH, 'r', encoding='utf-8') as f:
                self._schema = yaml.safe_load(f)
            if not isinstance(self._schema, dict) or 'paths' not in self._schema:
                raise CarapisClientError(f"Invalid schema format in {_SCHEMA_PATH}")
        except FileNotFoundError:
            raise CarapisClientError(f"Schema file not found at {_SCHEMA_PATH}")
        except yaml.YAMLError as e:
            raise CarapisClientError(f"Error parsing schema file {_SCHEMA_PATH}: {e}") from e
        except Exception as e:
            raise CarapisClientError(f"An unexpected error occurred loading schema: {e}") from e

        # Extract only endpoints matching the specified api_base_path
        self._endpoints = self._extract_endpoints(self.api_base_path)

    @staticmethod
    def _get_version() -> str:
        """Reads the version from __init__.py"""
        try:
            init_path = os.path.join(_SCRIPT_DIR, '__init__.py')
            with open(init_path, 'r', encoding='utf-8') as f:
                for line in f:
                    if line.startswith('__version__'):
                        # Execute the line in a temporary dict to get the value
                        local_dict = {}
                        exec(line, globals(), local_dict)
                        return local_dict.get('__version__', 'unknown')
        except Exception:
            pass  # Ignore errors reading version
        return 'unknown'

    def _extract_endpoints(self, base_path_prefix: str) -> Dict[str, Dict[str, Any]]:
        """Extracts endpoints starting with the specified base path from the schema."""
        endpoints = {}
        paths = self._schema.get('paths', {})
        for path, path_item in paths.items():
            # Ensure we only process paths under the specific API base path
            if path.startswith(base_path_prefix):
                # Store the path relative to the api_base_path
                relative_path = path[len(base_path_prefix):]
                if not relative_path.startswith('/'):  # Ensure leading slash
                    relative_path = '/' + relative_path

                for method, operation in path_item.items():
                    # Process only valid HTTP methods
                    if method.lower() in ['get', 'post', 'put', 'delete', 'patch', 'options', 'head']:
                        if 'operationId' in operation:
                            op_id = operation['operationId']
                            # Map operationId to method, path, and parameters definition
                            endpoints[op_id] = {
                                'method': method.upper(),
                                'path_template': relative_path,
                                'parameters': operation.get('parameters', [])
                            }
                        else:
                            # Warn if an operation within the target path is missing an ID
                            print(f"Warning: Missing operationId for {method.upper()} {path}")
        return endpoints

    def _request(self, method: str, endpoint_path_template: str,
                 path_params: Optional[Dict[str, Any]] = None,
                 query_params: Optional[Dict[str, Any]] = None,
                 json_data: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """Helper method to make authenticated requests."""
        # Substitute path parameters
        formatted_path = endpoint_path_template
        if path_params:
            try:
                # Ensure all necessary path params are present before formatting
                required_path_vars = [p.strip('{}') for p in endpoint_path_template.split('/') if p.startswith('{')]
                missing_vars = [var for var in required_path_vars if var not in path_params]
                if missing_vars:
                    raise CarapisClientError(f"Missing required path parameters {missing_vars} for endpoint {endpoint_path_template}")
                formatted_path = endpoint_path_template.format(**path_params)
            except KeyError as e:
                # This catch might be redundant now due to the check above, but keep for safety
                raise CarapisClientError(f"Missing path parameter key {e} during formatting for endpoint {endpoint_path_template}") from e

        # Construct the full URL using the class's api_base_path
        url = f"{self.base_url}{self.api_base_path}{formatted_path}"

        # Filter out None values from query_params before sending
        actual_query_params = {k: v for k, v in query_params.items() if v is not None} if query_params else None

        try:
            response = requests.request(
                method,
                url,
                headers=self._headers,
                params=actual_query_params,
                json=json_data,
                timeout=30
            )
            response.raise_for_status()
            # Handle cases where response might be empty (e.g., 204 No Content)
            if response.status_code == 204 or not response.content:
                return {}  # Return empty dict for no content
            return response.json()
        except requests.exceptions.HTTPError as e:
            try:
                error_details = response.json()
            except requests.exceptions.JSONDecodeError:
                error_details = response.text  # Use raw text if JSON decoding fails
            raise CarapisClientError(f"HTTP error {response.status_code} for {method} {url}: {error_details}") from e
        except requests.exceptions.RequestException as e:
            raise CarapisClientError(f"Request failed for {method} {url}: {e}") from e
        except ValueError as e:  # Catches JSONDecodeError if response.json() fails above
            raise CarapisClientError(f"Failed to decode JSON response for {method} {url}: {e}") from e

    def _prepare_params(self, operation_id: str, func_args: Dict[str, Any]) -> tuple[Dict[str, Any], Dict[str, Any]]:
        """Separates function arguments into path and query parameters based on schema."""
        if operation_id not in self._endpoints:
            raise CarapisClientError(f"Unknown operationId: {operation_id}")

        endpoint_info = self._endpoints[operation_id]
        param_defs = endpoint_info.get('parameters', [])

        path_params = {}
        query_params = {}
        known_param_names = set()

        for param_def in param_defs:
            param_name = param_def['name']
            param_in = param_def['in']
            is_required = param_def.get('required', False)
            known_param_names.add(param_name)

            if param_name in func_args:
                value = func_args[param_name]
                # Add to appropriate dictionary if value is not None
                # For required params, _call_endpoint will handle missing value check later if needed
                if value is not None:
                    if param_in == 'path':
                        path_params[param_name] = value
                    elif param_in == 'query':
                        query_params[param_name] = value
                    # Ignore other 'in' types like 'header' or 'cookie' for now
            elif is_required:
                # Raise error immediately if a required parameter is missing (and wasn't passed as None)
                raise CarapisClientError(f"Missing required parameter '{param_name}' for operation '{operation_id}'")

        # Warn about any extra arguments passed that are not in the schema definition
        extra_args = set(func_args.keys()) - known_param_names
        if extra_args:
            print(f"Warning: Unexpected arguments provided for '{operation_id}': {extra_args}")

        return path_params, query_params

    def _call_endpoint(self, operation_id: str, **kwargs) -> Dict[str, Any]:
        """Internal method to prepare and make the API call."""
        if operation_id not in self._endpoints:
            raise CarapisClientError(f"Unknown operationId: {operation_id}")

        endpoint_info = self._endpoints[operation_id]
        method = endpoint_info['method']
        path_template = endpoint_info['path_template']

        # Prepare path and query parameters from kwargs based on schema
        path_params, query_params = self._prepare_params(operation_id, kwargs)

        # Make the request
        # Pass path_params and query_params directly
        return self._request(method, path_template, path_params=path_params, query_params=query_params)

    # --- Explicit Endpoint Methods --- >
    # Signatures are now based on parameters defined in schema.yaml

    def list_dealers(self,
                     limit: Optional[int] = None,
                     ordering: Optional[str] = None,
                     page: Optional[int] = None,
                     search: Optional[str] = None,
                     type: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_business_dealers_list """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_business_dealers_list', **params)

    def get_dealer(self, user_id: str) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_business_dealers_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_business_dealers_retrieve', **params)

    def list_diagnosis_centers(self, dealer: Optional[str] = None,
                               limit: Optional[int] = None,
                               ordering: Optional[str] = None,
                               page: Optional[int] = None,
                               search: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_business_diagnosis_centers_list """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_business_diagnosis_centers_list', **params)

    def get_diagnosis_center(self, code: str) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_business_diagnosis_centers_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_business_diagnosis_centers_retrieve', **params)

    def list_manufacturers(self, country: Optional[str] = None,
                           limit: Optional[int] = None,
                           ordering: Optional[str] = None,
                           page: Optional[int] = None,
                           search: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_manufacturers_list """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_manufacturers_list', **params)

    def get_manufacturer(self, code: str) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_manufacturers_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_manufacturers_retrieve', **params)

    def get_manufacturer_stats(self) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_manufacturers_stats_retrieve """
        # No parameters defined in schema for this operation
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_manufacturers_stats_retrieve', **params)

    def list_model_groups(self, limit: Optional[int] = None,
                          manufacturer: Optional[str] = None,
                          ordering: Optional[str] = None,
                          page: Optional[int] = None,
                          search: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_model_groups_list """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_model_groups_list', **params)

    def get_model_group(self, code: str) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_model_groups_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_model_groups_retrieve', **params)

    def list_models(self, limit: Optional[int] = None,
                    model_group: Optional[str] = None,
                    ordering: Optional[str] = None,
                    page: Optional[int] = None,
                    search: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_models_list """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_models_list', **params)

    def get_model(self, code: str) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_catalog_models_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_catalog_models_retrieve', **params)

    def list_vehicles(self, body_type: Optional[str] = None, color: Optional[str] = None,
                      fuel_type: Optional[str] = None, grade: Optional[str] = None,
                      has_accidents: Optional[bool] = None, has_repairs: Optional[bool] = None,
                      limit: Optional[int] = None, manufacturer: Optional[str] = None,
                      max_mileage: Optional[int] = None, max_price: Optional[int] = None,
                      max_year: Optional[int] = None, min_mileage: Optional[int] = None,
                      min_price: Optional[int] = None, min_year: Optional[int] = None,
                      model: Optional[str] = None,
                      model_group: Optional[str] = None,
                      ordering: Optional[str] = None,
                      page: Optional[int] = None, search: Optional[str] = None,
                      transmission: Optional[str] = None, vehicle_id: Optional[int] = None,
                      vehicle_no: Optional[str] = None, vin: Optional[str] = None,
                      warranty_type: Optional[str] = None) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_vehicles_list """
        # Collect all parameters passed to the function
        params = locals()
        params.pop('self')  # Remove 'self' from parameters

        # Prepare arguments for _call_endpoint, potentially renaming model_group
        call_kwargs = {}
        for key, value in params.items():
            if key == 'model_group' and value is not None:
                # Map the function argument 'model_group' to the query parameter 'model__model_group'
                call_kwargs['model__model_group'] = value
            elif value is not None:
                # Pass other non-None parameters as is
                call_kwargs[key] = value

        # Important: Pass the modified kwargs to _call_endpoint
        return self._call_endpoint('encar_v2_vehicles_list', **call_kwargs)

    def get_vehicle(self, vehicle_id: int) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_vehicles_retrieve """
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_vehicles_retrieve', **params)

    def get_vehicle_enums(self) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_vehicles_enums_retrieve """
        # No parameters defined in schema
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_vehicles_enums_retrieve', **params)

    def get_vehicle_stats(self) -> Dict[str, Any]:
        """ Corresponds to operationId: encar_v2_vehicles_stats_retrieve """
        # No parameters defined in schema
        params = locals()
        params.pop('self')
        # No kwargs to handle
        return self._call_endpoint('encar_v2_vehicles_stats_retrieve', **params)

    # --- Helper Methods --- >

    def get_endpoint_details(self, operation_id: str) -> Optional[Dict[str, Any]]:
        """ Returns the schema details for a given operationId. """
        return self._endpoints.get(operation_id)

    def list_available_operations(self) -> List[str]:
        """ Lists all available operationIds loaded from the schema. """
        return list(self._endpoints.keys())



import os
import sys
import json
import shutil
import logging
from dotenv import load_dotenv
from typing import Optional, Any, Dict, Tuple
from pathlib import Path
from encar import CarapisClient, CarapisClientError

# --- Setup Path --- >
THIS_DIR = Path(__file__).parent
DOWNLOADS_DIR = THIS_DIR / "downloads"

load_dotenv()


class EncarApiRunner:
    """Runs a sequence of Carapis API calls, logging results and saving outputs."""

    def __init__(self, api_key: str):
        self.api_key = api_key
        self.script_dir = os.path.dirname(os.path.abspath(__file__))
        self.downloads_dir = DOWNLOADS_DIR
        self.logger = self._setup_logging()
        self.client = self._initialize_client()
        self._cleanup_downloads()

    def _setup_logging(self) -> logging.Logger:
        """Configures and returns a logger instance."""
        log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

        # Console Handler
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        logger.addHandler(console_handler)

        # File Handler (optional, uncomment to enable file logging)
        # log_file = os.path.join(self.script_dir, 'run_encar.log')
        # file_handler = logging.FileHandler(log_file, encoding='utf-8')
        # file_handler.setFormatter(log_formatter)
        # logger.addHandler(file_handler)

        return logger

    def _initialize_client(self) -> CarapisClient:
        """Initializes and returns the CarapisClient."""

        if not self.api_key:
            self.logger.error("CARAPIS_API_KEY is not set.")
            sys.exit(1)
        try:
            client = CarapisClient(api_key=self.api_key)
            self.logger.info(f"CarapisClient initialized for base URL: {client.base_url}")
            return client
        except CarapisClientError as e:
            self.logger.fatal(f"Failed to initialize CarapisClient: {e}")
            sys.exit(1)
        except Exception as e:
            self.logger.fatal(f"Unexpected error during client initialization: {e}")
            sys.exit(1)

    def _cleanup_downloads(self):
        """Removes and recreates the downloads directory."""
        self.logger.info(f"Cleaning up downloads directory: {self.downloads_dir}")
        try:
            if os.path.exists(self.downloads_dir):
                shutil.rmtree(self.downloads_dir)
                self.logger.info("Removed existing downloads directory.")
            os.makedirs(self.downloads_dir, exist_ok=True)
            self.logger.info(f"Recreated downloads directory: {self.downloads_dir}")
        except OSError as e:
            self.logger.error(f"Could not clean/create downloads directory {self.downloads_dir}: {e}")
            sys.exit(1)

    def _save_result(self, method_name: str, args: Dict, result: Any):
        """Saves the successful API result to a JSON file."""
        if not isinstance(result, dict) or not result:
            self.logger.debug(f"Skipping save for {method_name}: No valid dictionary result.")
            return

        try:
            filename_parts = [method_name]
            for key_arg in ['vehicle_id', 'code', 'user_id']:
                if key_arg in args and args[key_arg]:
                    filename_parts.append(str(args[key_arg]))

            filename = "_".join(filename_parts) + ".json"
            filepath = os.path.join(self.downloads_dir, filename)

            self.logger.info(f"Saving result of '{method_name}' to: {filepath}")
            with open(filepath, 'w', encoding='utf-8') as f:
                json.dump(result, f, ensure_ascii=False, indent=2)
        except Exception as e:
            self.logger.error(f"Failed to save result for {method_name} to file: {e}")

    def _run_business_endpoints(self) -> Tuple[Optional[str], Optional[str]]:
        """Runs API calls related to dealers and diagnosis centers."""
        self.logger.info("--- Running Business Endpoints ---")
        fetched_dealer_id = None
        fetched_diag_center_code = None
        res = None  # Initialize res

        # List Dealers
        method_name = "list_dealers"
        limit_arg, ordering_arg = 2, "name"
        args_dict = {"limit": limit_arg, "ordering": ordering_arg}  # For saving
        self.logger.info(f"Calling: {method_name} with limit={limit_arg}, ordering='{ordering_arg}'")
        try:
            # Explicit arguments
            res = self.client.list_dealers(limit=limit_arg, ordering=ordering_arg)
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res)
            if res and res.get('results') and len(res['results']) > 0:
                fetched_dealer_id = res['results'][0].get('user_id')
                self.logger.info(f"---> Fetched dealer_id: {fetched_dealer_id}")
            else:
                self.logger.warning("---> Could not fetch dealer_id from list_dealers.")
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        # Get Dealer (if ID fetched)
        if fetched_dealer_id:
            method_name = "get_dealer"
            args_dict = {"user_id": fetched_dealer_id}  # For saving
            self.logger.info(f"Calling: {method_name} with user_id='{fetched_dealer_id}'")
            try:
                # Explicit argument
                res_dealer = self.client.get_dealer(user_id=fetched_dealer_id)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_dealer)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping get_dealer - No ID fetched.")

        # List Diagnosis Centers
        method_name = "list_diagnosis_centers"
        limit_arg = 2
        args_dict = {"limit": limit_arg}  # For saving
        self.logger.info(f"Calling: {method_name} with limit={limit_arg}")
        res = None  # Reset res
        try:
            # Explicit argument
            res = self.client.list_diagnosis_centers(limit=limit_arg)
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res)
            if res and res.get('results') and len(res['results']) > 0:
                fetched_diag_center_code = res['results'][0].get('code')
                self.logger.info(f"---> Fetched diagnosis center code: {fetched_diag_center_code}")
            else:
                self.logger.warning("---> Could not fetch diagnosis center code from list_diagnosis_centers.")
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        # Get Diagnosis Center (if code fetched)
        if fetched_diag_center_code:
            method_name = "get_diagnosis_center"
            args_dict = {"code": fetched_diag_center_code}  # For saving
            self.logger.info(f"Calling: {method_name} with code='{fetched_diag_center_code}'")
            try:
                # Explicit argument
                res_diag = self.client.get_diagnosis_center(code=fetched_diag_center_code)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_diag)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping get_diagnosis_center - No code fetched.")

        return fetched_dealer_id, fetched_diag_center_code

    def _run_catalog_endpoints(self) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """Runs API calls related to manufacturers, model groups, and models."""
        self.logger.info("--- Running Catalog Endpoints ---")
        fetched_manufacturer_code = None
        fetched_model_group_code = None
        fetched_model_code = None
        res = None  # Initialize res

        # List Manufacturers
        method_name = "list_manufacturers"
        limit_arg = 2
        args_dict = {"limit": limit_arg}  # For saving
        self.logger.info(f"Calling: {method_name} with limit={limit_arg}")
        try:
            # Explicit argument
            res = self.client.list_manufacturers(limit=limit_arg)
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res)
            if res and res.get('results') and len(res['results']) > 0:
                fetched_manufacturer_code = res['results'][0].get('code')
                self.logger.info(f"---> Fetched manufacturer code: {fetched_manufacturer_code}")
            else:
                self.logger.warning("---> Could not fetch manufacturer code from list_manufacturers.")
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        if fetched_manufacturer_code:
            # Get Manufacturer
            method_name = "get_manufacturer"
            args_dict = {"code": fetched_manufacturer_code}  # For saving
            self.logger.info(f"Calling: {method_name} with code='{fetched_manufacturer_code}'")
            try:
                # Explicit argument
                res_mfr = self.client.get_manufacturer(code=fetched_manufacturer_code)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_mfr)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

            # List Model Groups
            method_name = "list_model_groups"
            limit_arg, ordering_arg = 2, "name"
            args_dict = {"manufacturer": fetched_manufacturer_code, "limit": limit_arg, "ordering": ordering_arg}  # For saving
            res_mg = None
            self.logger.info(f"Calling: {method_name} with manufacturer='{fetched_manufacturer_code}', limit={limit_arg}, ordering='{ordering_arg}'")
            try:
                # Explicit arguments
                res_mg = self.client.list_model_groups(manufacturer=fetched_manufacturer_code, limit=limit_arg, ordering=ordering_arg)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_mg)
                if res_mg and res_mg.get('results') and len(res_mg['results']) > 0:
                    fetched_model_group_code = res_mg['results'][0].get('code')
                    self.logger.info(f"---> Fetched model group code: {fetched_model_group_code}")
                else:
                    self.logger.warning("---> Could not fetch model group code from list_model_groups.")
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping manufacturer-dependent catalog calls - No manufacturer code fetched.")

        # Get Manufacturer Stats (Known to fail currently)
        method_name = "get_manufacturer_stats"
        args_dict = {}  # For saving
        self.logger.info(f"Calling: {method_name} (no args)")
        try:
            # No arguments
            res_mfr_stats = self.client.get_manufacturer_stats()
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res_mfr_stats)
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        if fetched_model_group_code:
            # Get Model Group
            method_name = "get_model_group"
            args_dict = {"code": fetched_model_group_code}  # For saving
            self.logger.info(f"Calling: {method_name} with code='{fetched_model_group_code}'")
            try:
                # Explicit argument
                res_get_mg = self.client.get_model_group(code=fetched_model_group_code)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_get_mg)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

            # List Models
            method_name = "list_models"
            limit_arg = 2
            args_dict = {"model_group": fetched_model_group_code, "limit": limit_arg}  # For saving
            res_mdl = None
            self.logger.info(f"Calling: {method_name} with model_group='{fetched_model_group_code}', limit={limit_arg}")
            try:
                # Explicit arguments
                res_mdl = self.client.list_models(model_group=fetched_model_group_code, limit=limit_arg)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_mdl)
                if res_mdl and res_mdl.get('results') and len(res_mdl['results']) > 0:
                    fetched_model_code = res_mdl['results'][0].get('code')
                    self.logger.info(f"---> Fetched model code: {fetched_model_code}")
                else:
                    self.logger.warning("---> Could not fetch model code from list_models.")
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping model group-dependent catalog calls - No model group code fetched.")

        if fetched_model_code:
            # Get Model
            method_name = "get_model"
            args_dict = {"code": fetched_model_code}  # For saving
            self.logger.info(f"Calling: {method_name} with code='{fetched_model_code}'")
            try:
                # Explicit argument
                res_get_mdl = self.client.get_model(code=fetched_model_code)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_get_mdl)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping get_model - No model code fetched.")

        return fetched_manufacturer_code, fetched_model_group_code, fetched_model_code

    def _run_vehicle_endpoints(self, fetched_manufacturer_code: Optional[str], fetched_model_group_code: Optional[str]):
        """Runs API calls related to vehicles."""
        self.logger.info("--- Running Vehicle Endpoints ---")
        fetched_vehicle_id = None
        res_list = None  # Initialize

        # List Vehicles (simple, to get an ID)
        method_name = "list_vehicles_for_id"
        limit_arg, ordering_arg = 1, "-created_at"
        args_dict = {"limit": limit_arg, "ordering": ordering_arg}  # For saving
        self.logger.info(f"Calling: {method_name} with limit={limit_arg}, ordering='{ordering_arg}'")
        try:
            # Explicit arguments
            res_list = self.client.list_vehicles(limit=limit_arg, ordering=ordering_arg)
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res_list)
            if res_list and res_list.get('results') and len(res_list['results']) > 0:
                fetched_vehicle_id = res_list['results'][0].get('vehicle_id')
                self.logger.info(f"---> Fetched vehicle_id: {fetched_vehicle_id}")
            else:
                self.logger.warning("---> Could not fetch vehicle_id from list_vehicles.")
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        # --- Additional List Examples ---
        if fetched_manufacturer_code:
            method_name_base = "list_vehicles_by_manufacturer"
            method_name_save = f"list_vehicles_by_{fetched_manufacturer_code}"  # Filename
            limit_arg = 2
            args_dict = {"limit": limit_arg, "manufacturer": fetched_manufacturer_code}  # For saving
            self.logger.info(f"Calling: {method_name_base} with limit={limit_arg}, manufacturer='{fetched_manufacturer_code}'")
            try:
                # Explicit arguments
                res_mfr_list = self.client.list_vehicles(limit=limit_arg, manufacturer=fetched_manufacturer_code)
                self.logger.info(f"SUCCESS: {method_name_base} for {fetched_manufacturer_code}")
                self._save_result(method_name_save, args_dict, res_mfr_list)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name_base} for {fetched_manufacturer_code} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name_base} for {fetched_manufacturer_code} - {type(e).__name__}: {e}")

        if fetched_manufacturer_code and fetched_model_group_code:
            # NOTE: model_group filter is now handled by the client library
            method_name_base = "list_vehicles_by_manufacturer_model_group"
            method_name_save = f"list_vehicles_{fetched_manufacturer_code}_{fetched_model_group_code}"  # Filename
            limit_arg = 2
            args_dict = {"limit": limit_arg, "manufacturer": fetched_manufacturer_code, "model_group": fetched_model_group_code}  # For saving
            self.logger.info(f"Attempting {method_name_base} with limit={limit_arg}, manufacturer='{fetched_manufacturer_code}', model_group='{fetched_model_group_code}'")
            try:
                # Explicit arguments
                res_mfr_mg_list = self.client.list_vehicles(limit=limit_arg, manufacturer=fetched_manufacturer_code, model_group=fetched_model_group_code)
                self.logger.info(f"SUCCESS: {method_name_base} for {fetched_manufacturer_code}/{fetched_model_group_code}")
                self._save_result(method_name_save, args_dict, res_mfr_mg_list)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name_base} for {fetched_manufacturer_code}/{fetched_model_group_code} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name_base} for {fetched_manufacturer_code}/{fetched_model_group_code} - {type(e).__name__}: {e}")

        method_name = "list_vehicles_year_2022"
        limit_arg, min_year_arg, max_year_arg = 2, 2022, 2022
        args_dict = {"limit": limit_arg, "min_year": min_year_arg, "max_year": max_year_arg}  # For saving
        self.logger.info(f"Calling: {method_name} with limit={limit_arg}, min_year={min_year_arg}, max_year={max_year_arg}")
        try:
            # Explicit arguments
            res_year_list = self.client.list_vehicles(limit=limit_arg, min_year=min_year_arg, max_year=max_year_arg)
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res_year_list)
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        # --- < Additional List Examples ---

        # Get Vehicle (if ID fetched)
        if fetched_vehicle_id:
            method_name = "get_vehicle"
            args_dict = {"vehicle_id": fetched_vehicle_id}  # For saving
            self.logger.info(f"Calling: {method_name} with vehicle_id={fetched_vehicle_id}")
            try:
                # Explicit argument
                res_vehicle = self.client.get_vehicle(vehicle_id=fetched_vehicle_id)
                self.logger.info(f"SUCCESS: {method_name}")
                self._save_result(method_name, args_dict, res_vehicle)
            except CarapisClientError as e:
                self.logger.error(f"FAILED (API Error): {method_name} - {e}")
            except Exception as e:
                self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")
        else:
            self.logger.warning("Skipping get_vehicle - No ID fetched.")

        # Get Vehicle Enums
        method_name = "get_vehicle_enums"
        args_dict = {}  # For saving
        self.logger.info(f"Calling: {method_name} (no args)")
        try:
            # No arguments
            res_enums = self.client.get_vehicle_enums()
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res_enums)
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

        # Get Vehicle Stats
        method_name = "get_vehicle_stats"
        args_dict = {}  # For saving
        self.logger.info(f"Calling: {method_name} (no args)")
        try:
            # No arguments
            res_stats = self.client.get_vehicle_stats()
            self.logger.info(f"SUCCESS: {method_name}")
            self._save_result(method_name, args_dict, res_stats)
        except CarapisClientError as e:
            self.logger.error(f"FAILED (API Error): {method_name} - {e}")
        except Exception as e:
            self.logger.error(f"FAILED (Unexpected Error): {method_name} - {type(e).__name__}: {e}")

    def run(self):
        """Executes the full sequence of API calls."""
        self.logger.info("=== Starting Encar API Run ===")
        _, _ = self._run_business_endpoints()
        mfr_code, mg_code, _ = self._run_catalog_endpoints()
        self._run_vehicle_endpoints(mfr_code, mg_code)
        self.logger.info("=== Finished Encar API Run ===")


def main():
    api_key = os.environ.get("CARAPIS_API_KEY")  # Removed default key

    if not api_key:
        # Logger is not initialized yet, use print for this critical error
        print("CRITICAL: CARAPIS_API_KEY environment variable not set.")
        print("Alternatively, export the environment variable.")
        sys.exit(1)

    try:
        runner = EncarApiRunner(api_key)
        runner.run()
    except Exception as e:
        # Catch any unexpected error during runner instantiation or run
        # Use print as logger might not be fully set up if error is early
        print(f"\nUNEXPECTED CRITICAL ERROR: {type(e).__name__}: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()
